from migen import *

from litex.gen import LiteXModule, ClockDomain, ClockSignal

from litex.soc.cores.cpu import vexriscv
from litex.soc.integration.soc_core import SoCCore
from litex.soc.integration.builder import Builder, builder_argdict, builder_args
from litex.soc.integration.soc_core import soc_core_argdict, soc_core_args
from litex.build.lattice.trellis import trellis_argdict, trellis_args

from litedram.modules import M12L64322A
from litedram.phy import GENSDRPHY, HalfRateGENSDRPHY
from liteeth.phy.ecp5rgmii import LiteEthPHYRGMII

from platform.colorlight_5a_75b_8_0 import Groovy1Platform
from hub75 import Hub75Driver, make_hub75_iodevice, Hub75VerilogDriver

class GroovySoC(SoCCore):
    def __init__(self, platform, sys_clk_freq,
                 use_spi = False,
                 **kwargs):

        SoCCore.__init__(self, platform, sys_clk_freq, ident="LiteX SoC for GroovyLight", **kwargs)


        self.crg = platform.get_crg(sys_clk_freq)
        self.submodules += self.crg
        print(kwargs)

        if not self.integrated_main_ram_size:
            self.sdrphy = HalfRateGENSDRPHY(platform.request("sdram"), sys_clk_freq)
            self.add_sdram("sdram",
                phy                     = self.sdrphy,
                module                  = M12L64322A(sys_clk_freq, "1:2"),
                l2_cache_size           = kwargs.get("l2_size", 8192),
            )


        self.submodules.ethphy = LiteEthPHYRGMII(
                clock_pads= self.platform.request("eth_clocks", 0),
                pads = self.platform.request("eth", 0),
                tx_delay = 0e-9, # not sure what this is
                )
        self.add_csr("ethphy")
        self.add_etherbone(phy=self.ethphy, ip_address="192.168.0.36", mac_address = 0x10e2d5000001, data_width=32)
        if use_spi:
            from litespi.modules import W25Q32JV as SpiFlashModule
            from litespi.opcodes import SpiNorFlashOpCodes
            self.mem_map["spiflash"] = 0x20000000
            mod = SpiFlashModule(SpiNorFlashOpCodes.READ_1_1_1)
            self.add_spi_flash(mode="1x", module=SpiFlashModule, with_master=False)
        self.platform.add_extension(make_hub75_iodevice(0, "j8"))
        hub_io = self.platform.request("hub75_iodev", 0)
        self.submodules.hub75 = hub75 = ClockDomainsRenamer("hub")(Hub75VerilogDriver())
        self.comb += [
                hub_io.r0.eq(hub75.o_panel_rgb0[0]),
                hub_io.g0.eq(hub75.o_panel_rgb0[1]),
                hub_io.b0.eq(hub75.o_panel_rgb0[2]),
                hub_io.r1.eq(hub75.o_panel_rgb1[0]),
                hub_io.g1.eq(hub75.o_panel_rgb1[1]),
                hub_io.b1.eq(hub75.o_panel_rgb1[2]),
                hub_io.clk.eq(hub75.o_display_clk),
                hub_io.addr.eq(hub75.o_addr_out),
                hub_io.oe.eq(hub75.o_out_enable),
                hub_io.stb.eq(hub75.o_latch),
        ]
        platform.add_source("verilog/hub75e.sv")






import argparse

def main():
    parser = argparse.ArgumentParser(description="Groovylight builder")

    builder_arggroup = parser.add_argument_group("builder options")
    soc_arggroup = parser.add_argument_group('SoC core options')
    soc_arggroup.add_argument("--with-spiflash", action="store_true", help="Use built in SPI flash")

    builder_args(builder_arggroup)
    soc_core_args(soc_arggroup)
    trellis_args(parser.add_argument_group('Trellis options'))
    args = parser.parse_args()
    platform = Groovy1Platform()
    soc = GroovySoC(platform, 75e6, **soc_core_argdict(args))


    builder = Builder(soc, **builder_argdict(args))

    builder.build()


if __name__ == "__main__":
    main()
